#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#include "config.h"
#include "imapfilter.h"
#include "account.h"
#include "filter.h"
#include "struct.h"


extern account_t *accounts;
extern mboxgrp_t *mboxgrps;

filter_t *filters = NULL;	/* First node of filters tree. */

filter_t *cur_fltr = NULL;	/* Current filter. */


void init_filter(filter_t * node);

void init_mask(mask_t * node);
void convert_date(mask_t * node);

void link_mbox_filter(filter_t * cf, mboxgrp_t * cg);

void string_upper(char *str, size_t size);


/*
 * Set new filter's variables to safe values.
 */
void
init_filter(filter_t * node)
{
	node->left = node->right = NULL;
	node->key[0] = '\0';
	node->mode = FILTER_MODE_AND;
	node->action.type = node->action.msgflags = 0;
	node->action.raccount = NULL;
	node->action.destmbox[0] = node->action.args[0] = '\0';
	node->masks = NULL;
	node->masknum = node->masklen = 0;
}


/*
 * A filter entry was declared, create it and set it's variables accordingly.
 */
int
set_filter(char *line, regmatch_t * m)
{
	filter_t *node;

	if (cur_fltr != NULL && cur_fltr->action.type == 0)
		return ERROR_CONFIG_PARSE;

	node = (filter_t *) xmalloc(sizeof(filter_t));

	init_filter(node);

	strncat(node->key, line + m[1].rm_so,
	    min(m[1].rm_eo - m[1].rm_so, KEY_LEN - 1));

	if (m[2].rm_so != -1) {
		if (!strncasecmp(line + m[2].rm_so + 1, "or", 2))
			node->mode = FILTER_MODE_OR;
		else
			node->mode = FILTER_MODE_AND;
	}
#ifdef DEBUG
	fprintf(stderr, "debug: FILTER: '%s' %s\n", node->key,
	    (node->mode == FILTER_MODE_OR ? "OR" : "AND"));
#endif

	INSERT_TREE(filters, node, filter);

	cur_fltr = node;

	return 0;
}


/*
 * Assign an action to the last declared filter.
 */
int
set_action(char *line, regmatch_t * m)
{
	char *c, *cp, *t;
	account_t *a;

	if (cur_fltr == NULL)
		return ERROR_CONFIG_PARSE;

	if (!strncasecmp(line + m[1].rm_so, "delete", 6))
		cur_fltr->action.type = FILTER_ACTION_DELETE;
	else if (!strncasecmp(line + m[1].rm_so, "copy", 4)) {
		cur_fltr->action.type = FILTER_ACTION_COPY;
		if (*(line + m[2].rm_so) == '"' &&
		    *(line + m[2].rm_eo - 1) == '"')
			strncat(cur_fltr->action.destmbox,
			    line + m[2].rm_so + 1,
			    min(m[2].rm_eo - m[2].rm_so - 2,
				MBOX_NAME_LEN - 1));
		else
			strncat(cur_fltr->action.destmbox, line + m[2].rm_so,
			    min(m[2].rm_eo - m[2].rm_so, MBOX_NAME_LEN - 1));
	} else if (!strncasecmp(line + m[1].rm_so, "move", 4)) {
		cur_fltr->action.type = FILTER_ACTION_MOVE;
		if (*(line + m[3].rm_so) == '"' &&
		    *(line + m[3].rm_eo - 1) == '"')
			strncat(cur_fltr->action.destmbox,
			    line + m[3].rm_so + 1,
			    min(m[3].rm_eo - m[3].rm_so - 2,
				MBOX_NAME_LEN - 1));
		else
			strncat(cur_fltr->action.destmbox, line + m[3].rm_so,
			    min(m[3].rm_eo - m[3].rm_so, MBOX_NAME_LEN - 1));
	} else if (!strncasecmp(line + m[1].rm_so, "rcopy", 5)) {
		cur_fltr->action.type = FILTER_ACTION_RCOPY;
		for (a = accounts; a; a = a->next)
			if (!strncasecmp(line + m[4].rm_so, a->key,
				strlen(a->key)))
				cur_fltr->action.raccount = a;
		if (!cur_fltr->action.raccount)
			return ERROR_CONFIG_PARSE;

		if (*(line + m[5].rm_so) == '"' &&
		    *(line + m[5].rm_eo - 1) == '"')
			strncat(cur_fltr->action.destmbox,
			    line + m[5].rm_so + 1,
			    min(m[5].rm_eo - m[5].rm_so - 2,
				MBOX_NAME_LEN - 1));
		else
			strncat(cur_fltr->action.destmbox, line + m[5].rm_so,
			    min(m[5].rm_eo - m[5].rm_so, MBOX_NAME_LEN - 1));
	} else if (!strncasecmp(line + m[1].rm_so, "rmove", 5)) {
		cur_fltr->action.type = FILTER_ACTION_RMOVE;
		for (a = accounts; a; a = a->next)
			if (!strncasecmp(line + m[6].rm_so, a->key,
				strlen(a->key)))
				cur_fltr->action.raccount = a;
		if (!cur_fltr->action.raccount)
			return ERROR_CONFIG_PARSE;

		if (*(line + m[7].rm_so) == '"' &&
		    *(line + m[7].rm_eo - 1) == '"')
			strncat(cur_fltr->action.destmbox,
			    line + m[7].rm_so + 1,
			    min(m[7].rm_eo - m[7].rm_so - 2,
				MBOX_NAME_LEN - 1));
		else
			strncat(cur_fltr->action.destmbox, line + m[7].rm_so,
			    min(m[7].rm_eo - m[7].rm_so, MBOX_NAME_LEN - 1));
	} else if (!strncasecmp(line + m[1].rm_so, "flag", 4)) {
		if (!strncasecmp(line + m[8].rm_so, "replace", 7))
			cur_fltr->action.type = FILTER_ACTION_FLAG_REPLACE;
		else if (!strncasecmp(line + m[8].rm_so, "add", 3))
			cur_fltr->action.type = FILTER_ACTION_FLAG_ADD;
		else
			cur_fltr->action.type = FILTER_ACTION_FLAG_REMOVE;

		c = cp = (char *)malloc(m[9].rm_eo - m[9].rm_so + 1);
		xstrncpy(c, line + m[9].rm_so, m[9].rm_eo - m[9].rm_so);

		t = strtok_r(c, ",", &c);
		while (t) {
			if (!strcasecmp(t, "none")) {
				if (m[9].rm_eo - m[9].rm_so != strlen("none"))
					return ERROR_CONFIG_PARSE;
			} else if (!strcasecmp(t, "seen"))
				cur_fltr->action.msgflags |= MESSAGE_FLAG_SEEN;
			else if (!strcasecmp(t, "answered"))
				cur_fltr->action.msgflags |=
				    MESSAGE_FLAG_ANSWERED;
			else if (!strcasecmp(t, "flagged"))
				cur_fltr->action.msgflags |=
				    MESSAGE_FLAG_FLAGGED;
			else if (!strcasecmp(t, "deleted"))
				cur_fltr->action.msgflags |=
				    MESSAGE_FLAG_DELETED;
			else if (!strcasecmp(t, "draft"))
				cur_fltr->action.msgflags |= MESSAGE_FLAG_DRAFT;
			else
				return ERROR_CONFIG_PARSE;

			t = strtok_r(NULL, ",", &c);
		}
		xfree(cp);
	} else if (!strncasecmp(line + m[1].rm_so, "list", 4)) {
		cur_fltr->action.type = FILTER_ACTION_LIST;
	}
	if (m[10].rm_so != -1) {
		strncat(cur_fltr->action.args, line + m[10].rm_so,
		    min(m[10].rm_eo - m[10].rm_so, ARGS_LEN - 2));
		while ((c = strchr(cur_fltr->action.args, ',')))
			*c = ' ';
	}
#ifdef DEBUG
	fprintf(stderr, "debug: ACTION: %d '%s' '%s' %d '%s'\n",
	    cur_fltr->action.type,
	    (cur_fltr->action.raccount ? cur_fltr->action.raccount->key : ""),
	    cur_fltr->action.destmbox,
	    cur_fltr->action.msgflags,
	    cur_fltr->action.args);
#endif
	return 0;
}


/*
 * Set new mask's variables to safe values.
 */
void
init_mask(mask_t * node)
{
	node->next = NULL;
	node->body[0] = '\0';
	node->type = 0;
}


/*
 * A new mask entry was declared, create it and set it's
 * variables accordingly.
 */
int
set_mask(char *line, regmatch_t * m, int mmt)
{
	int n;
	mask_t *node;
	char *bp;

	if (cur_fltr == NULL)
		return ERROR_CONFIG_PARSE;

	node = (mask_t *) xmalloc(sizeof(mask_t));

	init_mask(node);

	bp = node->body;

	/* If specified set mask's type. */
	if (m[2].rm_so != -1 && cur_fltr->masks) {
		if (!strncasecmp(line + m[2].rm_so, "or", 2)) {
			node->type = MASK_TYPE_OR;
		} else
			node->type = MASK_TYPE_AND;
	}
	/* Add NOT if specified. */
	if (m[3].rm_so != -1) {
		n = min(m[3].rm_eo - m[3].rm_so,
		    MASK_BODY_LEN - (bp - node->body) - 1);
		xstrncpy(bp, line + m[3].rm_so, n);
		string_upper(bp, n);
		*(bp + n - 1) = ' ';	/* In case it's '\t'. */
		*(bp + n) = '\0';
		bp += n;
	}
	/* Keyword of the search key. */
	n = min(m[4].rm_eo - m[4].rm_so,
	    MASK_BODY_LEN - (bp - node->body) - 3);
	xstrncpy(bp, line + m[4].rm_so, n);
	string_upper(bp, n);
	*(bp + n) = '\0';
	bp += n;

	if (mmt != MASK_MATCH_1) {
		*(bp++) = ' ';

		if (mmt != MASK_MATCH_4 && *(line + m[5].rm_so) != '"')
			*(bp++) = '"';
		*bp = '\0';

		n = min(m[5].rm_eo - m[5].rm_so,
		    MASK_BODY_LEN - (bp - node->body) - 2);
		xstrncpy(bp, line + m[5].rm_so, n);
		*(bp + n) = '\0';
		bp += n;

		if (mmt != MASK_MATCH_4 && *(line + m[5].rm_so) != '"')
			*(bp++) = '"';
		*bp = '\0';

		if (mmt == MASK_MATCH_3) {
			*(bp++) = ' ';

			if (*(line + m[6].rm_so) != '"')
				*(bp++) = '"';
			*bp = '\0';

			n = min(m[6].rm_eo - m[6].rm_so,
			    MASK_BODY_LEN - (bp - node->body) - 2);
			xstrncpy(bp, line + m[6].rm_so, n);
			*(bp + n) = '\0';
			bp += n;

			if (*(line + m[6].rm_so) != '"')
				*(bp++) = '"';
			*bp = '\0';
		}
		if (mmt == MASK_MATCH_4 && (strstr(node->body, "OLDER") ||
			strstr(node->body, "NEWER"))) {
			convert_date(node);
			bp = node->body + strlen(node->body);
		}
	}
	APPEND_LINKED_LIST(cur_fltr->masks, node, mask);

	cur_fltr->masknum++;
	cur_fltr->masklen += (bp - node->body);

#ifdef DEBUG
	fprintf(stderr, "debug: MASK: '%s'\n", node->body);
#endif

	return 0;
}


/*
 * Converts masks related to date filtering, because IMAP servers do not
 * understand for example "OLDER 3", but "BEFORE 18-Oct-2001" (if
 * hypothetically current date was 21-Oct-2001).
 */
void
convert_date(mask_t * node)
{
	char *cp, *c;
	char s[16];
	time_t te;
	struct tm *tl;

	cp = xstrdup(node->body);
	node->body[0] = '\0';

	if (strstr(cp, "NOT"))
		strncat(node->body, "NOT ", 4);

	if ((c = strstr(cp, "OLDER")))
		strncat(node->body, "BEFORE ", 7);
	else if ((c = strstr(cp, "NEWER")))
		strncat(node->body, "SINCE ", 6);

	c += 6;

	te = time(NULL) - (time_t) (strtoul(c, NULL, 10) * 24 * 60 * 60);
	tl = localtime(&te);

	if (strftime(s, 15, "%d-%b-%Y", tl))
		strncat(node->body, s, 15);

	xfree(cp);
}


/*
 * A new job was declared, link filters with mailbox-groups.
 */
int
set_job(char *line, regmatch_t * match)
{
	int n;
	char *ftok, *gtok, *fltr, *mbgrp, *f, *g;
	filter_t *cf;
	mboxgrp_t *cg;

	if (accounts == NULL || filters == NULL || cur_fltr->action.type == 0)
		return ERROR_CONFIG_PARSE;

	n = match[1].rm_eo - match[1].rm_so;
	fltr = (char *)xmalloc(n + 1);

	f = xstrncpy(fltr, line + match[1].rm_so, n);
	f[n] = '\0';

	n = match[2].rm_eo - match[2].rm_so;
	mbgrp = (char *)xmalloc(n + 1);

	/* Go through filters. */
	ftok = strtok_r(f, ",", &f);
	while (ftok != NULL) {
		FIND_TREE(filters, ftok, cf);
		if (cf == NULL)
			return ERROR_CONFIG_PARSE;

		g = xstrncpy(mbgrp, line + match[2].rm_so, n);
		g[n] = '\0';

		/* Go through mailbox groups. */
		gtok = strtok_r(g, ",", &g);
		while (gtok != NULL) {
			FIND_TREE(mboxgrps, gtok, cg);
			if (cg == NULL)
				return ERROR_CONFIG_PARSE;
			link_mbox_filter(cf, cg);

			gtok = strtok_r(NULL, ",", &g);
		}

		ftok = strtok_r(NULL, ",", &f);
	}

	xfree(fltr);
	xfree(mbgrp);

	return 0;
}


/*
 * Link a filter with a mailbox.
 */
void
link_mbox_filter(filter_t * cf, mboxgrp_t * cg)
{
	int i, j, f;

	for (i = 0; cg->mboxes[i] != NULL; i++) {
		for (f = j = 0; cg->mboxes[i]->filters[j] != NULL; j++)
			if (j == MBOX_FILTERS_MAX - 1 ||
			    !strcmp(cf->key, cg->mboxes[i]->filters[j]->key))
				f = 1;

		if (f)
			continue;

		cg->mboxes[i]->filters[j] = cf;
		cg->mboxes[i]->filters[j + 1] = NULL;

	}

#ifdef DEBUG
	fprintf(stderr, "debug: JOB: '%s' '%s'\n", cf->key, cg->key);
#endif
}


/*
 * Convert a string of specified size to upper case.
 */
void
string_upper(char *str, size_t size)
{
	unsigned int i;

	for (i = 0; i < size; i++, str++)
		*str = toupper(*str);
}
